#!/usr/bin/env python

# Copyright 2015 Earth Sciences Department, BSC-CNS

# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.

"""
Module containing functions to manage autosubmit's experiments.
"""
import os
import sys
import string
import time
import pickle
import datetime
import copy

import autosubmitAPIwu.database.db_common as db_common
from autosubmitAPIwu.job.job_list import JobList
from autosubmitAPIwu.job.job_package_persistence import JobPackagePersistence
from autosubmitAPIwu.job.job_grouping import JobGrouping
from autosubmitAPIwu.monitor.monitor import Monitor
from autosubmitAPIwu.job.job_common import Status, Type
from autosubmitAPIwu.experiment.statistics import ExperimentStats
from autosubmitAPIwu.config.basicConfig import BasicConfig
from autosubmitAPIwu.config.config_common import AutosubmitConfig
from bscearth.utils.config_parser import ConfigParserFactory
from bscearth.utils.log import Log
from bscearth.utils.date import date2str
from autosubmitAPIwu.autosubmit import Autosubmit

def get_experiment_stats(expid, filter_period, filter_type):
    """
    Lite version of the stats generator from Autosubmit autosubmit.py
    """
    error = False
    error_message = ""
    period_fi = ""
    period_ini = ""
    otherstats = list()
    jobs = list()
    results = None
    try:  
        if filter_period:
            filter_period = int(filter_period)                
        BasicConfig.read()
        as_conf = AutosubmitConfig(expid, BasicConfig, ConfigParserFactory())
        if not as_conf.check_conf_files():
            raise Exception('Can not run with invalid configuration')
        # notransitive as False by default
        job_list = Autosubmit.load_job_list(expid, as_conf, False)
        if filter_type:
            ft = filter_type
            if ft == 'Any':
                job_list = job_list.get_job_list()
            else:
                job_list = [job for job in job_list.get_job_list() if job.section == ft]
        else:
            ft = 'Any'
            job_list = job_list.get_job_list()

        period_fi = datetime.datetime.now().replace(second=0, microsecond=0)
        if filter_period and filter_period > 0:                     
            period_ini = period_fi - datetime.timedelta(hours=filter_period)
            job_list = [job for job in job_list if
                        job.check_started_after(period_ini) or job.check_running_after(period_ini)]
        else:
            period_ini = None
       
        if len(job_list) > 0:
            # Getting those predefined stats
            general_stats = []
            general_stats_path = os.path.join(BasicConfig.LOCAL_ROOT_DIR, expid, "tmp", expid + "_GENERAL_STATS")
            parser = AutosubmitConfig.get_parser(ConfigParserFactory(), general_stats_path)
            for section in parser.sections():
                general_stats.append((section, ''))
                general_stats += parser.items(section)
            otherstats = general_stats
            
            results = ExperimentStats(job_list, period_ini, period_fi)
            
            jobs = [job.name for job in results._jobs_list]
        else:
            raise Exception("There are no {0} jobs in the period from {1} to {2}...".format(ft, period_ini, period_fi))
        
    except Exception as e:
        error_message = str(e)
        error = True

    return {
        'error': error,
        'error_message': error_message,
        'periodo_fin': period_fi,
        'periodo_ini': period_ini,
        'otherstats': otherstats,
        'jobs': jobs,
        'stats': {            
            'max_time': results.max_time if results is not None else  "",
            'max_fail': results.max_fail if results is not None else  "",
            'queued': results.queued if results is not None else  "",
            'run': results.run if results is not None else  "",
            'failed_jobs': results.failed_jobs if results is not None else  "",
            'fail_queued': results.fail_queued if results is not None else  "",
            'fail_run': results.fail_run if results is not None else  "",
            'threshold': results.threshold if results is not None else  "",
            'totals': results._totals if results is not None else "",
        }
    }


def get_experiment_data(expid):
    """
    Get data of the experiment. Depends on the run cron of autosubmit.
    """
    path = 'Not found'
    conf_contents = ()
    tmp_contents = list()
    pkl_contents = list()
    plot_contents = list()
    status_contents = list()
    owner_id = 0
    owner_name = 'Not found'
    datetime_lastAccess = '1900-01-01 00:00:00'
    datetime_lastMod = '1900-01-01 00:00:00'
    error_message = 'None'
    contents = list()
    description = 'None'
    version = 'None'
    model = 'None'
    branch = 'None'
    hpc = 'None'
    updateTime = 10
    error = False
    running = False
    pkl_timestamp = 1000000
    reading = ""

    BasicConfig.read()
    path = BasicConfig.LOCAL_ROOT_DIR+'/'+expid

    if (os.path.exists(path)):
        try:  
            as_conf = AutosubmitConfig(expid, BasicConfig, ConfigParserFactory())
            if as_conf.check_conf_files() == False:
                raise Exception('Not a valid config file.')
            updateTime = as_conf.get_safetysleeptime()
            project_type = as_conf.get_project_type()
            if project_type != "none":
                if not as_conf.check_proj():
                    raise Exception('No valid proj conf found.')
            if (as_conf.get_svn_project_url()):
                model = as_conf.get_svn_project_url()
                branch = as_conf.get_svn_project_url()
            else:
                model = as_conf.get_git_project_origin()
                branch = as_conf.get_git_project_branch()
            if model is "":
                model = "Not Found"
            if branch is "":
                branch = "Not Found"

            submitter = Autosubmit._get_submitter(as_conf)
            hpc = as_conf.get_platform()
            
            resultBase = db_common.get_experiment_by_id(expid)          
            description = resultBase['description']
            version = resultBase['version']
            path_pkl = path + '/pkl'
            path_conf = path + '/conf'
            path_tmp = path + '/tmp'
            path_plot = path + '/plot'
            path_status = path + '/status'
            # print(path)
            contents = os.listdir(path)    
            # Getting pkl timestamp
            readingpkl = os.popen('ls -t ' + path_pkl + ' | grep ".pkl"').read() if (os.path.exists(path_pkl)) else ""
            #print(readingpkl)
            #pkl_contents = os.listdir(path_pkl) if (os.path.exists(path_pkl)) else list()
            #pkl_files = list(filter(lambda x: x.endswith(expid + '.pkl'), pkl_contents))
            if len(readingpkl) > 0:
                #print(readingpkl)
                pkl_file_name = readingpkl.split()[0]
                #print(pkl_file_name)
                current_stat = os.stat(path_pkl+'/'+pkl_file_name)
                pkl_timestamp = int(current_stat.st_mtime)
                #print(int(pkl_timestamp))

            status = os.stat(path)
            owner_id = status.st_uid
            owner_name = os.popen('id -nu ' + str(owner_id)).read().strip()     

            time_lastAccess = int(status.st_atime)
            datetime_lastAccess = datetime.datetime.utcfromtimestamp(time_lastAccess).strftime('%Y-%m-%d %H:%M:%S')
            time_lastMod = int(status.st_mtime)
            datetime_lastMod = datetime.datetime.utcfromtimestamp(time_lastMod).strftime('%Y-%m-%d %H:%M:%S')    
            size = str(status.st_size) + ' bytes' # get_size(path) # 

        except Exception as e:
            error_message = str(e)
            error = True
    else:
        error_message = str(expid) + " not found in " + str(BasicConfig.LOCAL_ROOT_DIR)

    result = {
     'expid' : expid,
     'path': path, 
     'contents': contents,           
     'contents_conf': conf_contents,
     'contents_tmp': tmp_contents, 
     'contents_pkl': pkl_contents,
     'contents_plot': plot_contents,
     'contents_status': status_contents,
    'owner_id': owner_id, 
    'owner': owner_name, 
    'time_last_access': datetime_lastAccess, 
    'time_last_mod': datetime_lastMod,
    'error_message': error_message,
    'description': description,
    'version' : version,
    'model': model,
    'branch': branch,
    'hpc': hpc,
    'updateTime': updateTime,
    'error': error,
    'running': running,
    'pkl_timestamp': pkl_timestamp}

    return result

def test_run(expid):
    """
    Tests if experiment is running
    """
    running = False
    error = False
    error_message = ""   

    try:
        BasicConfig.read()
        as_conf = AutosubmitConfig(expid, BasicConfig, ConfigParserFactory())
        if as_conf.check_conf_files() == False:
            raise Exception('Not a valid config file.')
        updateTime = as_conf.get_safetysleeptime()
        # print(str(expid) + " is Running?")
        pathlog = BasicConfig.LOCAL_ROOT_DIR + '/' + expid + '/' + BasicConfig.LOCAL_TMP_DIR + '/' + BasicConfig.LOCAL_ASLOG_DIR
        reading = os.popen('ls -t ' + pathlog + ' | grep "run.log"').read() if (os.path.exists(pathlog)) else ""
        if len(reading) > 0:
            log_file_name = reading.split()[0]
            current_stat = os.stat(pathlog+'/'+log_file_name)
            timest = current_stat.st_mtime
            #system_stat = os.stat(BasicConfig.LOCAL_ROOT_DIR)
            timesys = time.time()
            diff = timesys - timest
            print(diff)
            if (diff <= updateTime * 6):
                running = True
                return { 'running' : running,
                        'error' : error,
                        'error_message' : error_message }
        
        pathlog = BasicConfig.LOCAL_ROOT_DIR + '/' + expid + '/' + BasicConfig.LOCAL_TMP_DIR
        reading = os.popen('ls -t ' + pathlog + ' | grep "run.log"').read() if (os.path.exists(pathlog)) else ""
        if len(reading) > 0:
            log_file_name = reading.split()[0]
            current_stat = os.stat(pathlog+'/'+log_file_name)
            timest = current_stat.st_mtime
            #system_stat = os.stat(BasicConfig.LOCAL_ROOT_DIR)
            timesys = time.time()
            diff = timesys - timest
            print(diff)
            if (diff <= updateTime * 6):
                running = True
                return { 'running' : running,
                        'error' : error,
                        'error_message' : error_message }

    except Exception as ex:
        error = True
        error_message = str(ex)

    return { 'running' : running,
            'error' : error,
            'error_message' : error_message }


def get_experiment_run(expid):
    """
    Gets log content
    """
    # Initializing results:
    log_file_name = ""
    found = False
    log_file_lastmodified = ""
    timest = ""
    error = False
    error_message = ""    
    logcontent = []
    reading = ""

    try:
        BasicConfig.read()
        path = BasicConfig.LOCAL_ROOT_DIR + '/' + expid + '/' + BasicConfig.LOCAL_TMP_DIR + '/' + BasicConfig.LOCAL_ASLOG_DIR
        print(path)
        reading = os.popen('ls -t ' + path + ' | grep "run.log"').read() if (os.path.exists(path)) else ""

        # Finding log files    
        if len(reading) == 0:
            path = BasicConfig.LOCAL_ROOT_DIR + '/' + expid + '/' + BasicConfig.LOCAL_TMP_DIR
            reading = os.popen('ls -t ' + path + ' | grep "run.log"').read() if (os.path.exists(path)) else ""
            
        if len(reading) > 0:
            # run_logs.sort(reverse = True)
            log_file_name = reading.split()[0]
            print(log_file_name)
            current_stat = os.stat(path+'/'+log_file_name)
            timest = current_stat.st_mtime
            log_file_lastmodified = datetime.datetime.utcfromtimestamp(timest).strftime('%Y-%m-%d %H:%M:%S')
            found = True
            # line = subprocess.check_output(['tail', '-50', path+'/'+log_file_name])
            request = 'tail -50 ' + path + '/' + log_file_name
            print(request)
            last50 = os.popen(request)

            i = 0
            for item in last50.readlines():
                logcontent.append({'index': i, 'content': item[0:-1] })
                i+=1
    except Exception as e:
        error = True
        error_message = str(e)
        
    return { 
        'logfile' : log_file_name, 
        'found': found, 
        'lastModified': log_file_lastmodified, 
        'timeStamp': timest,
        'error': error,
        'error_message': error_message,
        'logcontent': logcontent }

def get_experiment_pkl(expid, modTimestamp):
    pkl_file_name = ""
    error = False
    error_message = ""
    has_changed = False
    pkl_content = list()
    timest = 1000000
    try:
        BasicConfig.read()
        path = BasicConfig.LOCAL_ROOT_DIR + '/' + expid + '/pkl'
        reading = os.popen('ls -t ' + path + ' | grep ".pkl"').read() if (os.path.exists(path)) else ""
        # pkl_contents = os.listdir(path) if (os.path.exists(path)) else list()
        # pkl_files = list(filter(lambda x: x.endswith(expid + '.pkl'), pkl_contents))
        if len(reading) > 0:
            pkl_file_name = reading.split()[0]
            current_stat = os.stat(path+'/'+pkl_file_name)
            timest = int(current_stat.st_mtime)
            #print(timest)
            if timest - int(modTimestamp) > 0:
                has_changed = True
        else:
            raise Exception('Empty pkl directory.')

        if has_changed == True:
            path_pkl = os.path.join(path, pkl_file_name)
            print(path_pkl)
            if os.path.exists(path_pkl):
                fd = open(path_pkl, 'r')            
                for item in pickle.load(fd):
                    pkl_content.append({'name': item[0],
                    'id': item[1],
                    'status_code': item[2],
                    'status': str(Status.VALUE_TO_KEY[item[2]]),
                    'status_color': Monitor.color_status(item[2]),
                    'priority': item[3]})
            else:
                raise Exception('File {0} does not exist'.format(path))

    except Exception as e:
        error = True
        error_message = str(e)

    return {
        'pkl_file_name' : pkl_file_name,
        'error' : error,
        'error_message' : error_message,
        'has_changed' : has_changed,
        'pkl_content' : pkl_content,
        'pkl_timestamp' : timest,
    }

def get_experiment_graph(expid):
    """
    Gets graph representation
    """
    base_list = dict()
    pkl_timestamp = 10000000
    try:
        notransitive = False
        print("Received " + str(expid))
        BasicConfig.read()
        exp_path = os.path.join(BasicConfig.LOCAL_ROOT_DIR, expid)
        tmp_path = os.path.join(exp_path, BasicConfig.LOCAL_TMP_DIR)
        as_conf = AutosubmitConfig(
            expid, BasicConfig, ConfigParserFactory())
        if not as_conf.check_conf_files():
            # Log.critical('Can not run with invalid configuration')
            raise Exception('Can not run with invalid configuration')


        job_list = Autosubmit.load_job_list(
            expid, as_conf, notransitive=notransitive)     

        job_list.sort_by_id()      
        base_list = job_list.get_graph_representation(BasicConfig)
    except Exception as e:
        return {'nodes': [], 
        'edges': [], 
        'fake_edges': [],
        'groups': [],
        'groups_data': {},
        'error': True, 
        'error_message': str(e), 
        'graphviz' : False,
        'max_children' : 0, 
        'max_parents' : 0,
        'total_jobs' : 0,
        'pkl_timestamp': 0}

    base_list['error'] = False
    base_list['error_message'] = 'None'
    base_list['pkl_timestamp'] = pkl_timestamp
    return base_list

def get_experiment_tree_structured(expid):
    """
    Current version of the tree visualization algorithm
    """
    base_list = dict()
    pkl_timestamp = 10000000
    try:
        notransitive = False
        print("Received Tree Request " + str(expid))
        BasicConfig.read()
        exp_path = os.path.join(BasicConfig.LOCAL_ROOT_DIR, expid)
        tmp_path = os.path.join(exp_path, BasicConfig.LOCAL_TMP_DIR)
        as_conf = AutosubmitConfig(
            expid, BasicConfig, ConfigParserFactory())
        if not as_conf.check_conf_files():
            # Log.critical('Can not run with invalid configuration')
            raise Exception('Can not run with invalid configuration')

        job_list = Autosubmit.load_job_list(
            expid, as_conf, notransitive=notransitive)     
        #job_list.sort_by_id()      
        #base_list['tree'] = job_list.get_tree_representation().traverse_format()
        
        tree_structure, current_collection, reference = job_list.get_tree_structured(BasicConfig)
        base_list['tree'] = tree_structure
        base_list['jobs'] = current_collection
        base_list['total'] = len(current_collection)
        base_list['reference'] = reference
    except Exception as e:
        return {'tree': [], 'jobs': [], 'total': 0, 'reference': [], 'error': True, 'error_message': str(e), 'pkl_timestamp': 0}
    base_list['error'] = False
    base_list['error_message'] = 'None'
    base_list['pkl_timestamp'] = pkl_timestamp
    return base_list
    


# def get_experiment_graph_grouped(expid, group_by, expand=list(), expand_status=list(), notransitive=False):
#     """
#     Old version of the grouping algorithm. Not used anymore.
#     """
#     nodes = list()
#     edges = list()
#     try:        
#         BasicConfig.read()
#         exp_path = os.path.join(BasicConfig.LOCAL_ROOT_DIR, expid)
#         as_conf = AutosubmitConfig(expid, BasicConfig, ConfigParserFactory())
#         if not as_conf.check_conf_files():            
#             raise Exception('Can not run with invalid configuration')
#         pkl_dir = os.path.join(BasicConfig.LOCAL_ROOT_DIR, expid, 'pkl')
#         job_list = Autosubmit.load_job_list(expid, as_conf, notransitive=notransitive,monitor=True)
#         jobs = job_list.get_job_list()

#         referenced_jobs_to_remove = set()
#         for job in jobs:
#             for child in job.children:
#                 if child not in jobs:
#                     referenced_jobs_to_remove.add(child)
#             for parent in job.parents:
#                 if parent not in jobs:
#                     referenced_jobs_to_remove.add(parent)

#         for job in jobs:
#             job.children = job.children - referenced_jobs_to_remove
#             job.parents = job.parents - referenced_jobs_to_remove

#         packages = JobPackagePersistence(os.path.join(BasicConfig.LOCAL_ROOT_DIR, expid, "pkl"),"job_packages_" + expid).load()


#         groups_dict = dict()
#         if group_by:
#             status = list()
#             if expand_status:
#                 for s in expand_status.split():
#                     status.append(Autosubmit._get_status(s.upper()))

#             job_grouping = JobGrouping(group_by, copy.deepcopy(jobs), job_list, expand_list=expand, expanded_status=status)
#             groups_dict = job_grouping.group_jobs()

#         monitor_exp = Monitor()
#         graph = monitor_exp.create_tree_list(expid, copy.deepcopy(jobs), packages, groups_dict, False)
#         #print("Honk")
#         exp = graph.get_subgraph_list()[0]
#         for u in exp.get_node_list():
#             print(u.obj_dict['name'])
        
#         for v in exp.get_edge_list():
#             print(str(v.get_source()) + " to " + str(v.get_destination()))
            
#         nodes = list()

#         edges_pre = set()
#         for v in exp.get_edge_list():
#             edges_pre.add((str(v.get_source()).replace('"','') , str(v.get_destination()).replace('"','')))
      
#         nodes_o = list()
#         nodes_dict = dict()
#         node_list = list()
#         for u in exp.get_node_list():
#             if u.obj_dict['name'] not in node_list:            
#                 nodes_dict[u.obj_dict['name']] = GroupNode(u.obj_dict['name'], 
#                                     u.obj_dict['attributes']['fillcolor'], 
#                                     list((y for x, y in edges_pre if x == u.obj_dict['name'])),
#                                     list((x for x, y in edges_pre if y == u.obj_dict['name'])),
#                                     'box' if u.obj_dict['attributes']['shape'] == 'box3d' else 'dot')
#                 node_list.append(u.obj_dict['name'])
#                 nodes_o.append(nodes_dict[u.obj_dict['name']])

#         for node in nodes_o:
#             node.update_dependencies(nodes_dict)


#         nodes_level = group_list_traverse_update(nodes_o)
#         edges = list()

#         for u,v in edges_pre:
#             edges.append({'from': str(u).replace('"',''), 'to': str(v).replace('"','')})

#         for node in nodes_level:            
#             nodes.append({'id' : node.name, 
#             'internal_id' : node.name, 
#             'label' : node.name,  
#             'shape': node.shape,
#             'level' : node.level,
#             'status_color' : node.color})


#     except Exception as ex:
#         return {'nodes': [], 'edges': [], 'error': True, 'error_message': str(ex)}

#     return {'nodes': nodes, 'edges': edges, 'error': False, 'error_message': 'None'}

# class GroupNode:
#     def __init__(self, name, color, children, parents, shape):
#         self.name = name
#         self.color = color
#         self._children = children
#         self.shape = shape
#         self._parents = parents
#         self.children = list()
#         self.parents = list()
#         self.level = 0
    
#     def has_parents(self):
#         return len(self.parents) > 0    
    
#     def update_dependencies(self, nodes_dict):
#         for parent in self._parents:
#             if parent != self.name:
#                 self.parents.append(nodes_dict[parent])
#         for child in self._children:
#             if child != self.name:
#                 self.children.append(nodes_dict[child])

# def group_list_traverse_update(node_list):
#     root = list()
#     ordered = list()
#     visited = list()
#     for node in node_list:
#         if node.has_parents() == False:
#             root.append(node)

#     for item in root:
#         #print("root: " + item.name)
#         _recursion_traverse_group(item, 0, visited, item, ordered)
#     return ordered

# def _recursion_traverse_group(node, level, visited, parent, ordered_list):
#     if node.name not in visited:
#         visited.append(node.name)
#         ordered_list.append(node)
#         level += 1
#         if len(node.children) > 0:
#             children = node.children
#             node.level = level              
#             for child in children:
#                 # print("Continues Recursion: " + str(child.name))
#                 # Continues recursion
#                 if child.name != node.name and child.name not in node._parents:
#                     print(node.name + " to " + child.name)
#                     _recursion_traverse_group(child, level, visited, node, ordered_list)
#         else:
#             node.level = level 
#     else:
#         if node.level <= parent.level:
#             level = parent.level + 1
#             node.level = level
#             children = node.children
#             for child in children:
#                 if child.name != node.name and child.name not in node._parents:
#                     print(node.name + " to " + child.name)
#                     _recursion_traverse_group(child, level, visited, node, ordered_list)

# def _job_running_check(job):
#     """
#     Receives a job and returns the data from its TOTAL_STATS file in an ordered way.
#     :param job: Job to be tested
#     :type job: Job object
#     :return: submit time, start time, end time, status
#     :rtype: 4-tuple
#     """
#     _tmp_path = "/esarchive/autosubmit/a2d0/tmp"
#     name = "a2d0_20161226_001_124_ARCHIVE"
#     values = list()
#     submit_time = 0
#     start_time = 0
#     finish_time = 0
#     current_status = "UNKNOWN"
#     path = os.path.join(_tmp_path, name + '_TOTAL_STATS')
#     if os.path.exists(path):
#         request = 'tail -1 ' + path
#         print(request)
#         last_line = os.popen(request).readline()
#         values = last_line.split(' ')
#     submit_time = values[0] if len(values) > 0 else 0
#     start_time = values[1] if len(values) > 1 else 0
#     finish_time = values[2] if len(values) > 2 else 0
#     current_status = values[3] if len(values) > 3 else "UNKNOWN"
    
#     return (submit_time, start_time, finish_time, current_status)

# def get_experiment_tree(expid):
#     """
#     Old confusing version
#     """
#     base_list = dict()
#     pkl_timestamp = 10000000
#     try:
#         notransitive = False
#         print("Received Tree Request " + str(expid))
#         BasicConfig.read()
#         exp_path = os.path.join(BasicConfig.LOCAL_ROOT_DIR, expid)
#         tmp_path = os.path.join(exp_path, BasicConfig.LOCAL_TMP_DIR)
#         as_conf = AutosubmitConfig(
#             expid, BasicConfig, ConfigParserFactory())
#         if not as_conf.check_conf_files():
#             # Log.critical('Can not run with invalid configuration')
#             raise Exception('Can not run with invalid configuration')


#         job_list = Autosubmit.load_job_list(
#             expid, as_conf, notransitive=notransitive)     
#         job_list.sort_by_id()      
#         base_list['tree'] = job_list.get_tree_representation().traverse_format()
#     except Exception as e:
#         return {'tree': [], 'error': True, 'error_message': str(e), 'pkl_timestamp': 0}
#     base_list['error'] = False
#     base_list['error_message'] = 'None'
#     base_list['pkl_timestamp'] = pkl_timestamp
#     return base_list



# def get_experiment_graph_format_test(expid):
#     """
#     Some testing
#     """    
#     base_list = dict()
#     pkl_timestamp = 10000000
#     try:
#         notransitive = False
#         print("Received " + str(expid))
#         BasicConfig.read()
#         exp_path = os.path.join(BasicConfig.LOCAL_ROOT_DIR, expid)
#         tmp_path = os.path.join(exp_path, BasicConfig.LOCAL_TMP_DIR)
#         as_conf = AutosubmitConfig(
#             expid, BasicConfig, ConfigParserFactory())
#         if not as_conf.check_conf_files():
#             # Log.critical('Can not run with invalid configuration')
#             raise Exception('Can not run with invalid configuration')


#         job_list = Autosubmit.load_job_list(
#             expid, as_conf, notransitive=notransitive)     

#         job_list.sort_by_id()      
#         base_list = job_list.get_graph_representation(BasicConfig)
#     except Exception as e:
#         return {'nodes': [], 
#         'edges': [], 
#         'error': True, 
#         'error_message': str(e), 
#         'graphviz' : False,
#         'max_children' : 0, 
#         'max_parents' : 0,
#         'total_jobs' : 0,
#         'pkl_timestamp': 0}
#     name_to_id = dict()
#     list_nodes = list()
#     list_edges = list()
#     i = 0

#     for item in base_list['nodes']:
#         name_to_id[item['id']] = i
#         list_nodes.append(i)
#         i += 1
#     for item in base_list['edges']:
#         list_edges.append((name_to_id[item['from']], name_to_id[item['to']]))

#     return list_nodes, list_edges